/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/iso.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>
#include <avs/arr.h>
#include <avs/dv_util.h>

#define ERR_RETURN(A) ERRerror("iso",1,ERR_ORIG, A); return(0);
#define MAX_NAME_SIZE 1024

int FUNCiso (OMobj_id in, double level, OMobj_id out, OMobj_id param,
             int ncell_data, int *cell_data);

int DViso_update(OMobj_id elem_id)
{
	OMobj_id in, out, level_id, param;
	double  level;
	int stat, *cell_data, size;

	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	param = OMfind_subobj(elem_id, OMstr_to_name("nparam"), OM_OBJ_RW);
	level_id = OMfind_subobj(elem_id, OMstr_to_name("level"), OM_OBJ_RD);
	OMget_real_val(level_id, &level);
	cell_data = (int *)OMret_name_array_ptr(elem_id, OMstr_to_name("cell_data"),
						OM_GET_ARRAY_RD, &size, NULL);
	if (size == 0)
		cell_data = NULL;
	if (cell_data == NULL)
		size = 0;

	stat = FUNCiso(in, level, out, param,  size, cell_data);
	if (cell_data)
		ARRfree(cell_data);
	return(stat);
}

int FUNCiso (OMobj_id in, double level, OMobj_id out, OMobj_id param,
             int ncell_data, int *cell_data)
{
	int   i, nspace, nnodes, out_ncells_line, out_ncells_tri, cell_ndim;
	int   out_nnodes, data_type, size, conn_size_line, conn_size_tri, veclen, ncomp;
	int   *out_nlist_line, *out_nlist_tri, *min_node_list, *max_node_list, stat, null_flag;
	int   nset, *ncells, **cell_ind, count, set_count;
	char  name[MAX_NAME_SIZE];
	float  *out_coord, *t;
	char   *node_data;
	char   *mesh_info;
	double null_value;
	OMobj_id   cell_set, in_cell_set;

	out_nnodes = 0;
	out_nlist_line = out_nlist_tri = min_node_list = max_node_list = NULL;
	cell_ind = NULL;
	out_coord = t = NULL;
	node_data = NULL;

	if (FLDget_nnodes(in, &nnodes) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_nspace (in, &nspace) != 1) {
		ERR_RETURN("Error getting nspace");
	}
	if (FLDget_node_data_ncomp(in, &ncomp) != 1) {
		ERR_RETURN("Error getting ncomp");
	}

	/********************************/
	/*   Free pre-allocated arrays  */
	/********************************/
	if (FLDset_nnodes (out, 0) != 1) {
		ERR_RETURN("Error setting nnodes");
	}
	if (FLDset_ncell_sets (out, 0) != 1) {
		ERR_RETURN("Error setting ncell_sets");
	}
	if (FLDset_int (param, "nnodes", 0) != 1) {
		ERR_RETURN("Error setting out nnodes");
	}
	min_node_list = max_node_list = NULL;
	if (FLDset_array_int (param, "min_node", 0, min_node_list,
			      OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting min_node_list");
	}
	if (FLDset_array_int (param, "max_node", 0, max_node_list,
			      OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting max_node_list");
	}
	if (FLDset_int (param, "nnodes", 0) != 1) {
		ERR_RETURN("Error setting out nnodes");
	}
	t = NULL;
	if (FLDset_array_float (param, "tpar", 0, t, OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting tpar");
	}
	out_coord = NULL;
	if (FLDset_coord(out, out_coord, 0, OM_SET_ARRAY_FREE) != 1) {
		ERR_RETURN("Error setting out_coord");
	}


	if (nnodes == 0 || ncomp == 0) {
		return(1);
	}

	if (FLDget_node_data_veclen(in, 0, &veclen) !=1) {
		ERR_RETURN("Error getting veclen");
	}

	if (veclen != 1) {
		ERR_RETURN("component is not scalar, iso is not performed");
	}

	if (FLDget_node_data(in, 0, &data_type, &node_data,
			      &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get node data");
	}
	if (FLDget_node_null_data(in, 0, &null_flag, (char *)&null_value) != 1) {
		ERR_RETURN("cannot get null data");
	}
	stat = FLDget_mesh_info(in, &mesh_info);
	if (!stat) {
		ERR_RETURN("cannot create cell table");
	}
	if (ncell_data) {
		if (FLDget_ncell_sets (in, &nset) != 1) {
			ERR_RETURN("Error getting ncell_sets");
		}
		ncells = (int *)malloc(nset*sizeof(int));
		cell_ind = (int **)malloc(nset*sizeof(int *));
	}
	else {
		cell_ind = NULL;
		ncells = NULL;
	}

	OMpush_status_range(0, 90);

	stat = UTILisosurf(mesh_info, node_data, data_type, 
			   null_flag, (char *)&null_value, (double)level,
			   &out_nnodes, &min_node_list, &max_node_list,
			   &out_ncells_line, &out_nlist_line, &conn_size_line, 
			   &out_ncells_tri, &out_nlist_tri, &conn_size_tri, 
			   ncells, cell_ind);

	OMpop_status_range();

	if (stat != 1) {
		if (FLDset_ncell_sets (out, 0) != 1) {
			ERR_RETURN("Error setting ncell_sets");
		}
		if (node_data)
			ARRfree(node_data);
		FLDfree_mesh_info(mesh_info);
		ERR_RETURN("no isosurface created");
	}

	/*** OUTPUT FIELD ***/

	if (FLDset_nspace (out, nspace) != 1) {
		ERR_RETURN("Error setting nspace");
	}
	if (FLDset_nnodes (out, out_nnodes) != 1) {
		ERR_RETURN("Error setting nnodes");
	}
	if (FLDset_int (param, "nnodes", out_nnodes) != 1) {
		ERR_RETURN("Error setting out nnodes");
	}

	if (out_nnodes) {
		if (FLDget_coord(out, &out_coord, &size, OM_GET_ARRAY_WR) != 1) {
			ERR_RETURN("Error setting coordinate array");
		}
		if (FLDget_coord_units(in, name, MAX_NAME_SIZE) == 1) {
			if (FLDset_coord_units (out, name) != 1) {
				ERR_RETURN("Error setting units");
			}
		}
		if (FLDset_array_int (param, "min_node", out_nnodes, 
				      min_node_list, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting min_node_list");
		}
		if (FLDset_array_int (param, "max_node", out_nnodes,
				      max_node_list, OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting max_node_list");
		}
		if (FLDget_array_float (param, "tpar", &t, &size, OM_GET_ARRAY_WR) != 1) {
			ERR_RETURN("Error setting t array");
		}

		UTILget_param(out_nnodes, data_type, node_data,
			      level, min_node_list, max_node_list, t);

		OMstatus_check(95,NULL,&stat);

		if (stat) {
			if (FLDset_ncell_sets (out, 0) != 1) {
				ERR_RETURN("Error setting ncell_sets");
			}
			if (FLDset_int (param, "nnodes", 0) != 1) {
				ERR_RETURN("Error setting out nnodes");
			}
			if (t)
				ARRfree(t);
			if (out_coord)
				ARRfree(out_coord);
			if (node_data)
				ARRfree(node_data);
			if (min_node_list)
				ARRfree(min_node_list);
			if (max_node_list)
				ARRfree(max_node_list);
			FLDfree_mesh_info(mesh_info);
			ERR_RETURN("no isosurface created");
		}


		UTILeval_coord(mesh_info, out_nnodes,
			       min_node_list, max_node_list, t, out_coord);

		OMstatus_check(100,NULL,&stat);

		if (stat) {
			if (FLDset_ncell_sets (out, 0) != 1) {
				ERR_RETURN("Error setting ncell_sets");
			}
			if (FLDset_int (param, "nnodes", 0) != 1) {
				ERR_RETURN("Error setting out nnodes");
			}
			if (t)
				ARRfree(t);
			if (out_coord)
				ARRfree(out_coord);
			if (node_data)
				ARRfree(node_data);
			if (min_node_list)
				ARRfree(min_node_list);
			if (max_node_list)
				ARRfree(max_node_list);
			FLDfree_mesh_info(mesh_info);
			ERR_RETURN("no isosurface created");
		}
		if (t)
			ARRfree(t);
		if (out_coord)
			ARRfree(out_coord);
	}
	else {
		if (min_node_list)
			ARRfree(min_node_list);
		if (max_node_list)
			ARRfree(max_node_list);
	}
	set_count = 0;
	if (out_ncells_line) {
		if (FLDadd_cell_set(out, "Line") != 1) {
			ERR_RETURN("Error setting cell type");
		}
		if (FLDget_cell_set(out, set_count++, &cell_set) != 1) {
			ERR_RETURN("Error getting cell set");
		}
		if (FLDset_ncells(cell_set, out_ncells_line) != 1) {
			ERR_RETURN("Error setting ncells");
		}
		if (FLDset_node_connect(cell_set, out_nlist_line, conn_size_line,
					    OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting cell connect list");
		}
		if (ncell_data) {
			if (FLDset_cell_set(cell_set, "Cell_Data_Set") != 1) {
				ERR_RETURN("Error setting cell type");
			}
			if (FLDset_cell_data_ncomp (cell_set, 0) != 1) {
				ERR_RETURN("Error setting cell data ncomp");
			}
			if (FLDset_cell_data_ncomp (cell_set, ncell_data) != 1) {
				ERR_RETURN("Error setting cell data ncomp");
			}
			for (count=0,i=0; i<nset; i++) {
				if (FLDget_cell_set(in, i, &in_cell_set) != 1) {
					ERR_RETURN("Error getting cell set");
				}
				if (FLDget_cell_ndim(in_cell_set,  &cell_ndim) != 1) {
					ERR_RETURN("cannot get cell ndim");
				}
				if (cell_ndim != 2)
					continue;
				stat = DVcopy_cell_data (in, in_cell_set, cell_set, 
							 ncells[i], cell_ind[i], 
							 ncell_data, cell_data, count);
				if (stat != 1) {
					if (FLDset_cell_data_ncomp (cell_set, 0) != 1) {
						ERR_RETURN("Error setting cell data ncomp");
					}
					for(count=0; count<nset; count++)
						if (cell_ind[count])
							ARRfree(cell_ind[count]);
					break;
				}
				count += ncells[i];
				if (cell_ind[i]) {
					ARRfree(cell_ind[i]);
					cell_ind[i] =NULL;
				}
			}
		}
	}
	else {
		if (out_nlist_line)
			ARRfree(out_nlist_line);
	}
	if (out_ncells_tri) {
		if (FLDadd_cell_set(out, "Tri") != 1) {
			ERR_RETURN("Error setting cell type");
		}
		if (FLDget_cell_set(out, set_count, &cell_set) != 1) {
			ERR_RETURN("Error getting cell set");
		}
		if (FLDset_ncells(cell_set, out_ncells_tri) != 1) {
			ERR_RETURN("Error setting ncells");
		}
		if (FLDset_node_connect(cell_set, out_nlist_tri, conn_size_tri,
					    OM_SET_ARRAY_FREE) != 1) {
			ERR_RETURN("Error setting cell connect list");
		}
		if (ncell_data) {
			if (FLDset_cell_set(cell_set, "Cell_Data_Set") != 1) {
				ERR_RETURN("Error setting cell type");
			}
			if (FLDset_cell_data_ncomp (cell_set, 0) != 1) {
				ERR_RETURN("Error setting cell data ncomp");
			}
			if (FLDset_cell_data_ncomp (cell_set, ncell_data) != 1) {
				ERR_RETURN("Error setting cell data ncomp");
			}
			for (count=0,i=0; i<nset; i++) {
				if (FLDget_cell_set(in, i, &in_cell_set) != 1) {
					ERR_RETURN("Error getting cell set");
				}
				if (FLDget_cell_ndim(in_cell_set,  &cell_ndim) != 1) {
					ERR_RETURN("cannot get cell ndim");
				}
				if (cell_ndim != 3)
					continue;
				stat = DVcopy_cell_data (in, in_cell_set, cell_set, 
							 ncells[i], cell_ind[i], 
							 ncell_data, cell_data, count);
				if (stat != 1) {
					if (FLDset_cell_data_ncomp (cell_set, 0) != 1) {
						ERR_RETURN("Error setting cell data ncomp");
					}
					for(count=0; count<nset; count++)
						if (cell_ind[count])
							ARRfree(cell_ind[count]);
					break;
				}
				count += ncells[i];
				if (cell_ind[i]) {
					ARRfree(cell_ind[i]);
					cell_ind[i] = NULL;
				}
			}
		}
	}
	else {
		if (out_nlist_tri)
			ARRfree(out_nlist_tri);
	}
	if (ncells)
		free(ncells);
	if (cell_ind)
		free(cell_ind);
	if (node_data)
		ARRfree(node_data);
	FLDfree_mesh_info(mesh_info);

	return(1);
}
